//===========================================================================
//
// Mesh.C
//
// By Poman So
//
//===========================================================================

#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#define Mesh_MAX 100000000L
#include "Mesh.h"

#include <iostream>
using namespace std;

static void newfailed_handler(){
    cout.flush();
    cerr << "\n\nOut of free memory for mesh!\n";
    exit(1);
}//--- newfailed_handler ---//


Mesh::Mesh(VecU& vu, VecR& vr, Real ier, Real iur)
:Node(vr/vu, ier, iur){
    Uint n;
    long msize=vu.x*vu.y*vu.z*72*sizeof(Real);
    if (msize>Mesh_MAX){
        cout.flush();
        cerr << "\n\nMesh ("<<vu.x<<"x"<<vu.y<<"x"<<vu.z<<"):"<<msize<<
                " bytes is too large to be handled by the system!\n";
        exit(1);
    }
    nx=vu.x;  ny=vu.y;  nz=vu.z;  x=vr.x;  y=vr.y;  z=vr.z;
    nyz=ny*nz; nxyz=nx*nyz;
    for(n=0;n<N_Lines;n++) vp[n]=NULL;
    nNodes=0;  node_index=NULL;  nodes=NULL;  current_index=0;
    nb=0;  n1p=n2p=NULL;   refl=NULL;
}//--- Mesh::Mesh ---//


Mesh::~Mesh(){
    Uint n;
    if (vp[0]){
        delete node_index;
        for(n=0;n<N_Lines;n++) delete vp[n];
    }
    if (nNodes) delete nodes;
    if (nb){
        delete n1p;  delete n2p;  delete refl;
    }
}//--- Mesh::~Mesh ---//


Real Mesh::Cal_Dt(){
    Real dt=Node::Cal_Dt();
    int  i;
    for (i=0;i<nNodes;i++)
        dt = Min(dt, nodes[i].Cal_Dt());
    return dt;
}//--- Mesh::Cal_Dt ---//


void Mesh::GetMemory(){
//extern void (*_new_handler)();
//void (*original_handler)();
    Uint i;
//  original_handler = _new_handler;
//  _new_handler = newfailed_handler;
    if (vp[0]){
        delete node_index;
        for(i=0;i<N_Lines;i++) delete vp[i];
    }
    for(i=0;i<N_Lines;i++) vp[i] = new Real[nxyz];
    node_index = new char[nxyz];
    Init();
//  _new_handler = original_handler;
}//--- Mesh::GetMemory ---//


void Mesh::Init(){
    Real *v;
    Uint i, j;
    for (i=0;i<N_Lines;i++)
        for (v=vp[i], j=0;j<nxyz;j++) v[j]=0;
}//--- Mesh::Init ---//


fstream &Mesh::IO(fstream &fs){
    Uint i;
    for (i=0;i<N_Lines;i++)  DataIO(fs, vp[i], sizeof(Real)*nxyz);
    return fs;
}//--- Mesh::IO ---//

void Mesh::Scatter(){
    Uint i;
    Real *p1, *p2, *p3, *p4, *p5, *p6, *p7, *p8, *p9, *p10, *p11, *p12,
            *p13, *p14, *p15, *p16, *p17, *p18, **p;
    Real a, b, c, Jx, Jy, Jz, Jxv16, Jyv17, Jzv18, Kx, Ky, Kz, Ux, Uy, Uz, Wx, Wy, Wz;
    p = vp;
    p1=*p++;   p2=*p++;   p3=*p++;   p4 =*p++;  p5 =*p++;  p6 =*p++;
    p7=*p++;   p8=*p++;   p9=*p++;   p10=*p++;  p11=*p++;  p12=*p++;
    p13=*p++;  p14=*p++;  p15=*p++;  p16=*p++;  p17=*p++;  p18=*p++;
    for(i=0;i<nxyz;i++){
        if (node_index[i]!=current_index)
            *((Node*)this)=nodes[current_index=node_index[i]];
        a = *p1 + *p12;    b = *p2 + *p9;    Ux=a+b+ *p13;  Wx=b-a;
        a = *p3 + *p11;    b = *p4 + *p8;    Uy=a+b+ *p14;  Wy=b-a;
        a = *p5 + *p7;     b = *p6 + *p10;   Uz=a+b+ *p15;  Wz=b-a;
        a = *p4 - *p8;     b = *p5 - *p7;    Jx=b-a+ *p16;  Kx=a+b;
        a = *p6 - *p10;    b = *p2 - *p9;    Jy=b-a+ *p17;  Ky=a+b;
        a = *p1 - *p12;    b = *p3 - *p11;   Jz=b-a+ *p18;  Kz=a+b;
        *p13 = s1*Ux - *p13;
        *p14 = s2*Uy - *p14;
        *p15 = s3*Uz - *p15;
        *p16 += s4*Jx;
        *p17 += s5*Jy;
        *p18 += s6*Jz;
        Jxv16  = Jx + *p16++;
        Jyv17  = Jy + *p17++;
        Jzv18  = Jz + *p18++;
        a=Ux-(*p13++ +s7*Ux);
        b=Kz+Jzv18;  c=a+Wx;  *p1++=0.25*(c+b);  *p12++=0.25*(c-b);
        b=Ky-Jyv17;  c=a-Wx;  *p2++=0.25*(c+b);  *p9++ =0.25*(c-b);
        a=Uy-(*p14++ +s8*Uy);
        b=Kz-Jzv18;  c=a+Wy;  *p3++=0.25*(c+b);  *p11++=0.25*(c-b);
        b=Kx+Jxv16;  c=a-Wy;  *p4++=0.25*(c+b);  *p8++ =0.25*(c-b);
        a=Uz-(*p15++ +s9*Uz);
        b=Kx-Jxv16;  c=a+Wz;  *p5++=0.25*(c+b);  *p7++ =0.25*(c-b);
        b=Ky+Jyv17;  c=a-Wz;  *p6++=0.25*(c+b);  *p10++=0.25*(c-b);
    }
}//--- Mesh::Scatter ---//

void Mesh::Transfer(){
    Real *p1, *p2, *p3, *p4, *p5, *p6, *p7, *p8, *p9, *p10, *p11, *p12, **p;
    Uint i, j, k, ijk=0, ipjk=nyz, ijpk=nz, ijkp=1, nxm1=nx-1, nym1=ny-1, nzm1=nz-1;
    p = vp;
    p1=*p++;  p2=*p++;  p3=*p++;  p4 =*p++;  p5 =*p++;  p6 =*p++;
    p7=*p++;  p8=*p++;  p9=*p++;  p10=*p++;  p11=*p++;  p12=*p++;
    for(i=0;i<nx;i++)
        for(j=0;j<ny;j++)
            for(k=0;k<nz;k++, ijk++, ipjk++, ijpk++, ijkp++){
                if(i<nxm1){ExV(p10[ijk], p6[ipjk]);  ExV(p11[ijk], p3[ipjk]);}
                if(j<nym1){ExV( p7[ijk], p5[ijpk]);  ExV(p12[ijk], p1[ijpk]);}
                if(k<nzm1){ExV( p8[ijk], p4[ijkp]);  ExV( p9[ijk], p2[ijkp]);}
            }
}//--- Mesh::Transfer ---//


void Mesh::Reflect(Boolean forward){
    Real *p1, *p2, *p3, *p4, *p5, *p6, *p7, *p8, *p9, *p10, *p11, *p12, **p, *pz, *pzm, rc, t;
    Uint x1, x2, y1, y2, z1, z2, n, i, j, im, jm, dj, k;
    p = vp;
    p1=*p++;  p2=*p++;  p3=*p++;  p4 =*p++;  p5 =*p++;  p6 =*p++;
    p7=*p++;  p8=*p++;  p9=*p++;  p10=*p++;  p11=*p++;  p12=*p++;
    //--------------------
    // Short circuit stubs
    //--------------------
    for(i=15;i<18;i++)
        for(pz=vp[i], pzm=pz+nxyz;pz<pzm;pz++) *pz=-*pz;
    //-----------
    // Boundaries
    //-----------
    for(n=0;n<nb;n++){
        rc=(forward)?refl[n]:(refl[n]!=0.0)?1/refl[n]:0.0;
        // There is no good way to handle 1/0.  In fact, 1/refl could causes
        // instability during backward simulation.
        //
        x1=n1p[n].x;  y1=n1p[n].y; z1=n1p[n].z;
        x2=n2p[n].x;  y2=n2p[n].y; z2=n2p[n].z;
        if(x1==x2){ // yz-plane
            dj=z2-z1;
            if(x1>0 && x1<nx)
                for(i=x1*nyz+y1*nz, im=i+(y2-y1)*nz;i<im;i+=nz)
                    for(j=i+z1, k=j-nyz, jm=j+dj;j<jm;j++, k++){
                        t=p3[j]*rc; p3[j]=p11[k]*rc; p11[k]=t;
                        t=p6[j]*rc; p6[j]=p10[k]*rc; p10[k]=t;
                    }
            else if(x1==0 && rc!=1)
                for(i=y1*nz, im=y2*nz;i<im;i+=nz)
                    for(j=i+z1, jm=j+dj;j<jm;j++){p3[j]*=rc;  p6[j]*=rc;}
            else if(x1==nx && rc!=1)
                for(i=nxyz-nyz+y1*nz, im=i+(y2-y1)*nz;i<im;i+=nz)
                    for(j=i+z1, jm=j+dj;j<jm;j++){p10[j]*=rc;  p11[j]*=rc;}
        }
        else if(y1==y2){ // xz-plane
            dj=z2-z1;
            if(y1>0 && y1<ny)
                for(i=x1*nyz+y1*nz, im=i+(x2-x1)*nyz;i<im;i+=nyz)
                    for(j=i+z1, k=j-nz, jm=j+dj;j<jm;j++, k++){
                        t=p1[j]*rc; p1[j]=p12[k]*rc; p12[k]=t;
                        t=p5[j]*rc; p5[j]=p7[k]*rc;  p7[k]=t;
                    }
            else if(y1==0 && rc!=1)
                for(i=x1*nyz, im=x2*nyz;i<im;i+=nyz)
                    for(j=i+z1, jm=j+dj;j<jm;j++){p1[j]*=rc;  p5[j]*=rc;}
            else if(y1==ny && rc!=1)
                for(i=x1*nyz+nyz-nz, im=x2*nyz+nyz-nz;i<im;i+=nyz)
                    for(j=i+z1, jm=j+dj;j<jm;j++){p7[j]*=rc;  p12[j]*=rc;}
        }
        else{ // xy-plane
            dj=(y2-y1)*nz;
            if(z1>0 && z2<nz)
                for(i=x1*nyz+y1*nz+z1, im=i+(x2-x1)*nyz;i<im;i+=nyz)
                    for(j=i, k=j-1, jm=j+dj;j<jm;j+=nz, k+=nz){
                        t=p2[j]*rc; p2[j]=p9[k]*rc; p9[k]=t;
                        t=p4[j]*rc; p4[j]=p8[k]*rc; p8[k]=t;
                    }
            else if(z1==0 && rc!=1)
                for(i=x1*nyz+y1*nz, im=i+(x2-x1)*nyz;i<im;i+=nyz)
                    for(j=i, jm=j+dj;j<jm;j+=nz){p2[j]*=rc;  p4[j]*=rc;}
            else if(z1==nz && rc!=1)
                for(i=x1*nyz+y1*nz+nz-1, im=i+(x2-x1)*nyz;i<im;i+=nyz)
                    for(j=i, jm=j+dj;j<jm;j+=nz){p8[j]*=rc;  p9[j]*=rc;}
        }
    }
}//--- Mesh::Reflect --//


void Mesh::InjectEH(VecU& vu, VecR& E, VecR& H){
    Uint ijk=Addr(vu.x, vu.y, vu.z);
    Real uEx2=E.x, vEy2=E.y, wEz2=E.z, uZoHx2=-H.x, vZoHy2=-H.y, wZoHz2=-H.z;
    if (node_index[ijk]!=current_index)
        *((Node*)this)=nodes[current_index=node_index[ijk]];
    uEx2*=0.5*u*(4+yx+gx)/(4+yx);	uZoHx2*=0.5*u*ZO;
    vEy2*=0.5*v*(4+yy+gy)/(4+yy);	vZoHy2*=0.5*v*ZO;
    wEz2*=0.5*w*(4+yz+gz)/(4+yz);	wZoHz2*=0.5*w*ZO;
    vp[ 0][ijk] += uEx2+wZoHz2;
    vp[ 1][ijk] += uEx2-vZoHy2;
    vp[ 2][ijk] += vEy2-wZoHz2;
    vp[ 3][ijk] += vEy2+uZoHx2;
    vp[ 4][ijk] += wEz2-uZoHx2;
    vp[ 5][ijk] += wEz2+vZoHy2;
    vp[ 6][ijk] += wEz2+uZoHx2;
    vp[ 7][ijk] += vEy2-uZoHx2;
    vp[ 8][ijk] += uEx2+vZoHy2;
    vp[ 9][ijk] += wEz2-vZoHy2;
    vp[10][ijk] += vEy2+wZoHz2;
    vp[11][ijk] += uEx2-wZoHz2;
    if(yx!=0) vp[12][ijk] += uEx2*yx;
    if(yy!=0) vp[13][ijk] += vEy2*yy;
    if(yz!=0) vp[14][ijk] += wEz2*yz;
    if(zx!=0) vp[15][ijk] += -zx*uZoHx2;
    if(zy!=0) vp[16][ijk] += -zy*vZoHy2;
    if(zz!=0) vp[17][ijk] += -zz*wZoHz2;
}//--- Mesh::InjectEH ---//


void Mesh::InjectEH(VecU &p1, VecU &p2, Dist sx, Dist sy, Dist sz,
        VecR &E, VecR &H){
    Real f, CFX, CFY, CFZ;
    VecU vu;
    VecR fE, fH;
    CFX = PI/(p2.x - p1.x + 1);
    CFY = PI/(p2.y - p1.y + 1);
    CFZ = PI/(p2.z - p1.z + 1);
    for (vu.x=p1.x;vu.x<=p2.x;vu.x++)
        for (vu.y=p1.y;vu.y<=p2.y;vu.y++)
            for (vu.z=p1.z;vu.z<=p2.z;vu.z++){
                switch (sx){
                    case CONSTANT:	f = 1;				break;
                    case HALF_SINE:	f = sin((vu.x-p1.x+0.5)*CFX);	break;
                    case HALF_COSINE:	f = cos((vu.x-p1.x+0.5)*CFX);	break;
                }
                switch (sy){
                    case CONSTANT:	f *= 1;				break;
                    case HALF_SINE:	f *= sin((vu.y-p1.y+0.5)*CFY);	break;
                    case HALF_COSINE:	f *= cos((vu.y-p1.y+0.5)*CFY);	break;
                }
                switch (sz){
                    case CONSTANT:	f *= 1;				break;
                    case HALF_SINE:	f *= sin((vu.z-p1.z+0.5)*CFZ);	break;
                    case HALF_COSINE:	f *= cos((vu.z-p1.z+0.5)*CFZ);	break;
                }
                fE = f*E;
                fH = f*H;
                InjectEH(vu, fE, fH);
            }
}//--- Mesh::InjectEH ---//


void Mesh::InjectEH(VecR &p1, VecR &p2, Dist sx, Dist sy, Dist sz,
        VecR &E, VecR &H){
    VecU ijk, IJK;
    CtoN(p1, ijk);
    CtoN(p2, IJK);
    if (IJK.x>0 && IJK.x>ijk.x) IJK.x--;
    if (IJK.y>0 && IJK.y>ijk.y) IJK.y--;
    if (IJK.z>0 && IJK.z>ijk.z) IJK.z--;
    InjectEH(ijk, IJK, sx, sy, sz, E, H);
}//--- Mesh::InjectEH ---//


Real Mesh::GetField(VecU& vu, int fm){
    Uint ijk=Addr(vu.x, vu.y, vu.z);
    Real ft;
    if (node_index[ijk]!=current_index)
        *((Node*)this)=nodes[current_index=node_index[ijk]];
    switch(fm){
        case 1:
            ft = vp[0][ijk]+vp[1][ijk]+vp[8][ijk]+vp[11][ijk];
            return (yx==0)?ft/(2*u):2*(ft+vp[12][ijk])/(u*(4+yx+gx));
        case 2:
            ft = vp[2][ijk]+vp[3][ijk]+vp[7][ijk]+vp[10][ijk];
            return (yy==0)?ft/(2*v):2*(ft+vp[13][ijk])/(v*(4+yy+gy));
        case 4:
            ft = vp[4][ijk]+vp[5][ijk]+vp[6][ijk]+vp[ 9][ijk];
            return (yz==0)?ft/(2*w):2*(ft+vp[14][ijk])/(w*(4+yz+gz));
        case 8:
            ft = vp[3][ijk]-vp[4][ijk]+vp[ 6][ijk]-vp[ 7][ijk];
            return (zx==0)?-ft/(2*u*ZO):2*(vp[15][ijk]-ft)/(u*ZO*(4+zx));
        case 16:
            ft = vp[5][ijk]-vp[1][ijk]+vp[ 8][ijk]-vp[ 9][ijk];
            return (zy==0)?-ft/(2*v*ZO):2*(vp[16][ijk]-ft)/(v*ZO*(4+zy));
        case 32:
            ft = vp[0][ijk]-vp[2][ijk]+vp[10][ijk]-vp[11][ijk];
            return (zz==0)?-ft/(2*w*ZO):2*(vp[17][ijk]-ft)/(w*ZO*(4+zz));
    }
    return 0;
}//--- Mesh::GetField ---//


Real Mesh::GetField(VecR &vr, int fm){
    VecU vu;
    CtoN(vr, vu);
    if (vu.x>0) vu.x--;
    if (vu.y>0) vu.y--;
    if (vu.z>0) vu.z--;
    return GetField(vu, fm);
}//--- Mesh::GetFiels ---//


void Mesh::GetField(VecU &u, VecU &v, int fm, Real *f, Real mf){
    VecU t=u;
    Uint k=0;
    if (u.x==v.x)
        for (;t.y<=v.y;t.y++)
            for (t.z=u.z;t.z<=v.z;t.z++)
                f[k++]=mf*GetField(t, fm);
    else if (u.y==v.y)
        for (;t.z<=v.z;t.z++)
            for (t.x=u.x;t.x<=v.x;t.x++)
                f[k++]=mf*GetField(t, fm);
    else // (u.z==v.z)
        for (;t.x<=v.x;t.x++)
            for (t.y=u.y;t.y<=v.y;t.y++)
                f[k++]=mf*GetField(t, fm);
}//--- Mesh::GetField ---//


void Mesh::CtoN(VecR& vr, VecU& vu){
    Uint i, j, k;
    i=Uint(vr.x/u+0.5); if(i>nx)i=nx;
    j=Uint(vr.y/v+0.5); if(j>ny)j=ny;
    k=Uint(vr.z/w+0.5); if(k>nz)k=nz;
    vu=VecU(i, j, k);
}//--- Mesh::CtoN ---//


void Mesh::NtoC(VecU& vu, VecR& vr){
    vr=VecR(vu.x*u, vu.y*v, vu.z*w);
}//--- Mesh::NtoC ---//


void Mesh::SetNodes(Uint n, Float8 *f8){
    if (nNodes) delete nodes;
    if ((nNodes=n)>0){
        Uint i, x, y, z;
        float *fp, er, ur, dt;
        VecR  r1, r2;
        VecU  n1, n2;
        nodes = new Node[nNodes];
        for (fp=f8[0], i=0;i<nNodes;i++){
            r1.x=*fp++;  r1.y=*fp++;  r1.z=*fp++;
            r2.x=*fp++;  r2.y=*fp++;  r2.z=*fp++;
            er  =*fp++;  ur  =*fp++;
            CtoN(r1, n1); CtoN(r2, n2);
            nodes[i].SetErUr(er, ur);
            for (x=n1.x;x<n2.x;x++)
                for (y=n1.y;y<n2.y;y++)
                    for (z=n1.z;z<n2.z;z++)
                        node_index[Addr(x, y, z)]=i;
        }
        for (dt=Cal_Dt(), i=0;i<nNodes;i++) nodes[i].SMatrix(dt);
        *((Node*)this)=nodes[current_index=0];
    }
}//--- Mesh::SetNodes ---//


void Mesh::SetBR(Uint n, Float7 *f7){
    float *fp;
    VecR  f1p, f2p;
    if(n==0) return;
    if(nb){delete n1p;  delete n2p;  delete refl;}
    nb  = n;
    n1p = new VecU[nb];  n2p = new VecU[nb];  refl= new Real[nb];
    for(fp=f7[0], n=0;n<nb;n++){
        f1p.x=*fp++; f1p.y=*fp++; f1p.z=*fp++;
        f2p.x=*fp++; f2p.y=*fp++; f2p.z=*fp++;
        refl[n] =*fp++;
        CtoN(f1p, n1p[n]);  CtoN(f2p, n2p[n]);
    }
}//--- Mesh::SetBR ---//

//===========================================================================
//
// End of Mesh.C
//
//===========================================================================
